#ifndef SCTARCHIVE_IOMANAGERDB_H
#define SCTARCHIVE_IOMANAGERDB_H

#include "Sct/Archive/IOManagerArchive.h"
#include "Sct/IoExceptions.h"
#include "Sct/LogicErrors.h"
#include "SctData/ResultHeader.h"
#include <mysql.h>
#include "CalibrationController/IS/TestData.h"

#include <boost/date_time/posix_time/posix_time.hpp>
//#include "sqlplus.hh"

class Sct::Serializable;
//class MysqlQuery;

namespace SctArchiving {
  /** exception thrown when queries go wrong */
class MysqlException : public Sct::IoException{
public:
  MysqlException(const string& msg, const string& file, int line);
  MysqlException(const string& msg, const string& query, const string& file, int line);
  ~MysqlException() throw();
private:
  MysqlException();
  std::string query;
};

class ZlibStringCompressor;
  /**
     This IOManager reads from and writes to a database.
     It is expected to be used in archival, and so also makes use of streamer 
     versioning by inheriting from IOManagerStreamerVersioned.
     Uses a <A HREF="http://www.mysql.com">mysql</a> back-end.
   */
class IOManagerDB : public Sct::Archive::IOManagerArchive {
public:
  /** 
      retrieve access to singleton 
  */
  static IOManagerDB& instance();
  /** 
      @{ @name IOManager stuff
      write using default streamers 
  */
  virtual void write(const Sct::Serializable& ob, const Sct::IOParams* params = 0) const;
  /** 
      read using appropriate streamers 
      @}
  */
  virtual shared_ptr<Sct::Serializable> read(const string& name, const Sct::IOParams* params = 0) const;
  virtual ~IOManagerDB();
  /** 
      get some status information 
   */
  virtual std::string status() const;
  /** 
      Add a run - gets data set identifier from getDataSetID
      
  */
  virtual void addRun(unsigned long runNumber, bool trash) const;

  /** 
      Add a test @param alsoAddScans if this is true, it will loop over the scans in the test
      adding them too
   */
  virtual void addTest(const TestData& testdata, bool alsoAddScans) const;

  /**
      Add a scan. Identifies correct Test from the run# and the 1st scan number.
  */  
  virtual void addScan(const TestData& testdata, unsigned scanNumber) const;

  /**
      Add a module.
  */  
  virtual void addModule(const std::string& layerName,
                         const std::string& LMTnr,
                         const std::string& poistionNumber,
                         const std::string& moduleNumber) const;

  /**
      Add module-scan combination raw data table.
  */  
  virtual void addRawDataRecord(const std::string& moduleId, const std::string& scanId) const;

  /**
      Add binary data to raw data table.
  */  
  virtual void addRawDataBlob(const std::string& moduleId, const std::string& scanId,
                              const char *chBuf, const long bufSize) const;

  /**
      Get binary data from raw data table.
  */  
  virtual std::string getRawDataBlob(const std::string& moduleId,
                                     const std::string& scanId) const;

  /** 
      get string identifying data set e.g. Barrel3_Cold_Oxford 
  */
  virtual std::string getDataSetName() const;

  /** 
      set the string identifying data set e.g. Barrel3_Cold_Oxford
  */
  virtual void setDataSetName(const std::string& dataSetName);

  /** 
      conversion just removes 'T' from isotime 
   */
  static std::string convertToMysqlDateTimeFormat(const std::string& isotime);

  /**
     @{ @name methods for getting table row identifiers
     Use strings for now, since we are always communicating with sql 
     
     Return empty string if not found!

     Uses the current DataSetName
   */
  std::string getDataSetID() const;
  std::string getTestTypeID(const std::string& testName) const;
  std::string getRunID(const unsigned runNumber) const;
  std::string getTestID(const std::string& runId, const unsigned startScanNumber) const;
  std::string getScanID(const std::string& testId, const unsigned scanNumber) const;
  std::string getLayerID(const std::string& layerName) const;
  std::string getRowID(const std::string& layerId, const std::string& LMTnr) const;
  std::string getPositionID(const std::string& rowId, const std::string& positionNumber) const;
  std::string getModuleID(const std::string& moduleNumber) const;
  /**
     @}
   */
 private:
  /** throw MysqlException if there is an error in the mysql object */
  void checkForMySqlErrors(const string& msg, const string& file, int line) const;

  /** throw MysqlException if there is an error in the mysql object - listing query */
  void checkForMySqlErrors(const string& msg, const string& query, const string& file, int line) const;

  IOManagerDB();  ///< constructor.
  //MysqlQuery getQuery() const; ///< get a null query.
  static IOManagerDB* s_man; ///< pointer to singleton
  /**
     keep a string compressor handy to zip and unzip BLOB data.
  */
  boost::shared_ptr<ZlibStringCompressor> m_compressor;
  //AJB ONLY FOR MYSQL++ boost::shared_ptr<Connection> m_connection; ///< pointer to the database connection

  MYSQL* m_mysql; ///< the main MYSQL object
  
  /**
     get the header from a TestResult or ScanResult
  */
  static const SctData::ResultHeader& getHeader( const Sct::Serializable& ob);
  /**
     change query to prepare to insert ob.
  */
  //void prepareInsertion(const Sct::Serializable& ob, MysqlQuery& q) const;
  /**
     make the "`WHERE'" part of the query matching the name by building a 
     IOName from it.
  */
  //static std::string getQueryMatching(const string& name);
  /**
     find the name of the table which the ob should to into.
   */
  static std::string getTable(const Sct::Serializable& ob);
  /**
     name of data set  - e.g. barrel3ColdAtOxford
   */
  std::string m_dataSetName;
  /**
     gets result, using msql_free_result in deleter
   */
  boost::shared_ptr<MYSQL_RES> getResult() const;

  /**
     mutex for timing information
  */
  mutable boost::recursive_mutex timingAccess;
  
  mutable unsigned m_nRead; ///< total number got out of database  
  mutable unsigned m_nWrite; ///< total number put into database
  mutable unsigned m_nCompress; ///< # of strings compressed
  mutable unsigned m_nDecompress; ///< # of strings decompressed
  mutable unsigned m_nBlobAdd; ///< # of raw objects added to database
  mutable unsigned m_nQueries; ///< # of raw objects added to database

  mutable boost::posix_time::time_duration read_time; ///< time spent reading from database
  mutable boost::posix_time::time_duration write_time; ///< time spent writing to database
  mutable boost::posix_time::time_duration compress_time; ///< time spent compressing data
  mutable boost::posix_time::time_duration decompress_time; ///< time spent decompressing data
  mutable boost::posix_time::time_duration blob_add_time; ///< time spent adding raw data blobs to db
  mutable boost::posix_time::time_duration query_time; ///< time spent adding raw data blobs to db
  mutable boost::posix_time::time_duration ostream_time; ///< time spent adding raw data blobs to db
  mutable boost::posix_time::time_duration istream_time; ///< time spent adding raw data blobs to db


  class MySqlFreeResultDeleter{
  public:
    void operator() (MYSQL_RES*) throw();
  };  
  /**
     utility method for getting single id
     Or more precise, this method returns the result of a query deisgned to return one
     record (row) and one field (column). An exeption is raised if the query returns 
     0 or more than 1 records. Otherwise the method returns the contents of the first field,
     other fields are ignored.
   */
  std::string executeQueryReturnID(std::string query) const;
  void executeQuery(std::string query) const;
};
}
#endif
