// For stat
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <boost/lexical_cast.hpp>

#include "dalImpl.h"
#include "dalImplUtils.h"
#include "../commonUtils.h"

#include "Barrel.h"
#include "Channel.h"
#include "Crate.h"
#include "Geography.h"
#include "GeographyMUR.h"
#include "Partition.h"
#include "Power.h"
#include "PowerCard.h"
#include "PowerChannel.h"
#include "PowerCrate.h"
#include "PowerParam.h"
#include "Rod.h"
#include "RodMUR.h"
#include "RodModule.h"
#include "RodRModule.h"
#include "Row.h"
#include "SCT_Chip.h"
#include "SCT_Module.h"
#include "Slave.h"
#include "Tim.h"

#include "../configRegistry.h"

class DALFactory : public SctConfiguration::Factory {
  SctConfiguration::Configuration *defaultConfig() {
    const char *envname = "SCTDAQ_ROD_CONFIGURATION_PATH";
    char *envfilename = getenv(envname);
    return new ConfigurationDALImpl(envfilename);
  }
  SctConfiguration::Configuration *fromFile(std::string filename) {
    SctConfiguration::Configuration *newConfig = new ConfigurationDALImpl(filename);
    return newConfig;
  }
};

bool dalInMap = SctConfiguration::FactoryManager::instance().addFactory("DAL", *new DALFactory());

namespace {
  void parseRcFunction(std::string content, float *rc_params) {
    char *buffer = (char *)malloc(strlen(content.c_str()) + 1);
    strcpy(buffer, content.c_str());
    char *token = strtok(buffer, " ");
    int nextVal = -1;
    do {
      if(strcmp(token, "p0") == 0) {
	nextVal = 0;
      } else if(strcmp(token, "p1") == 0) {
	nextVal = 1;
      } else if(strcmp(token, "p2") == 0) {
	nextVal = 2;
      } else {
	if(nextVal < 0) {
	  std::cerr << "Bad rc function parameters " << token << std::endl;
	} else {
	  rc_params[nextVal] = strtod(token, NULL);
	}
	nextVal = -1;
      }
    } while((token = strtok(NULL, " ")) != 0);
  }
}

const std::string ConfigurationDALImpl::schemaName = "SCT.schema.xml";

bool ConfigurationDALImpl::createFile(std::string fileName) {
  std::list<std::string> includes;
  includes.push_back(schemaName);
  log << "Creating new Config file\n";
  if(conf().create("", fileName, includes) == false) {
    std::cerr << "ERROR: failed to create file \'" << fileName << "\'\n";
    conf().abort();
    return 0;
  } else {
    conf().commit();
    return 1;
  }
} 

ConfigurationDALImpl::ConfigurationDALImpl(std::string fileName) : m_conf(new ::Configuration("oksconfig")),
								   log(),
								   builder(new ConfDALUtils::Builder(*m_conf, *this, log)), 
								   lookup(new ConfDALUtils::LookUp(*m_conf, *this, log)) {
  log << "ConfigurationDALImpl constructor\n";

  struct stat fileStat;
  int statReturn = stat(fileName.c_str(), &fileStat);
  if(statReturn == 0) {
    std::cerr << "File: " << fileName << " already exists\n";
    if(fileStat.st_size == 0) {
      std::cerr << "Removing empty file in the way\n";
      unlink(fileName.c_str());

      createFile(fileName);
    } else {
      if(conf().load(fileName)) {
	std::cerr << "Loaded successfully\n";
	lastFilename = fileName;
      } else {
	throw "Something";
      }
    }
  } else {
    std::cerr << "File: " << fileName << " stat returned " << statReturn << "\n";

    if(createFile(fileName)) {
      lastFilename = fileName;
    }
  }

  std::list<std::string> returnIncludes;
  
  conf().get_includes(fileName, returnIncludes);

  if(find(returnIncludes.begin(), returnIncludes.end(), schemaName) == returnIncludes.end()) {
    std::cerr << "Adding schema as include\n";
    conf().add_include(fileName, schemaName);
  }

  conf().commit();
}

ConfigurationDALImpl::~ConfigurationDALImpl() {
  delete builder;
  delete m_conf;
}

std::list<unsigned int> ConfigurationDALImpl::listPartitions() {
  std::vector<const SCTDAL::Partition *> partitions;
  conf().get<SCTDAL::Partition>(partitions);

  std::list<unsigned int> result;
  for(SCTDAL::PartitionIterator i = partitions.begin(); i != partitions.end(); i++) {
    result.push_back((*i)->get_id());
  }

  return result;
}

std::list<unsigned int> ConfigurationDALImpl::listCratesInPartition(unsigned int partition) {
  const SCTDAL::Partition *partOb = lookup->getPartition(partition);

  std::vector<const SCTDAL::Crate *> crates = partOb->get_crates();

  std::list<unsigned int> result;
  for(SCTDAL::CrateIterator i = crates.begin(); i != crates.end(); i++) {
    result.push_back((*i)->get_id());
  }

  return result;
}

bool ConfigurationDALImpl::isDummyCrate(unsigned int partition, unsigned int crate) {
#warning "Obsolete"
  std::cerr << "Obsolete at " << __LINE__ << std::endl;

  return false;
}

std::list<unsigned int> ConfigurationDALImpl::listRodsInCrate(unsigned int partition, unsigned int crate) {
  const SCTDAL::Crate *crateOb = lookup->getCrate(partition, crate);

  std::vector<const SCTDAL::Rod *> rods = crateOb->get_rods();

  std::list<unsigned int> result;
  for(SCTDAL::RodIterator i = rods.begin(); i != rods.end(); i++) {
    result.push_back((*i)->get_id());
  }

  return result;
}

std::list<unsigned int> ConfigurationDALImpl::listMURSInRod(unsigned int partition, unsigned int crate, unsigned int rod) {
  const SCTDAL::Rod *rodObj = lookup->getRod(partition, crate, rod);
  if(!rodObj) {
    throw SctConfiguration::ConfigurationException("No Rod");
  }

  std::list<unsigned int> result;

  std::vector<const SCTDAL::RodMUR *> rodMurs = rodObj->get_rodMURs();

  for(SCTDAL::RodMURIterator i = rodMurs.begin(); i != rodMurs.end(); i++) {
    const SCTDAL::RodMUR *rodMur = *i;

    result.push_back(rodMur->get_id());
  }

  return result;
}

std::list<std::string> ConfigurationDALImpl::listModulesInMUR(unsigned int partition, unsigned int MUR) {
  const SCTDAL::RodMUR *murObj = lookup->getMur(MUR);

  if(!murObj) {
    throw SctConfiguration::ConfigurationException("No MUR to get Modules");
  }

  std::vector<const SCTDAL::RodModule *> modules = murObj->get_rodModules();

  std::list<std::string> result;
  for(SCTDAL::RodModuleIterator i = modules.begin(); i!=modules.end(); i++) {
    const SCTDAL::SCT_Module *modObj = (*i)->get_module();
    if(modObj) {
      result.push_back(modObj->get_sn());
    } else {
      std::cerr << "RodModule points to non-existant module\n";
    }
  }

  return result;
}

std::list<std::string> ConfigurationDALImpl::listRModulesInMUR(unsigned int partition, unsigned int MUR) {
  const SCTDAL::RodMUR *murObj = lookup->getMur(MUR);

  if(!murObj) {
    throw SctConfiguration::ConfigurationException("No MUR to get R modules");
  }

  std::vector<const SCTDAL::RodRModule *> modules = murObj->get_rodRModules();

  std::list<std::string> result;
  for(SCTDAL::RodRModuleIterator i = modules.begin(); i!=modules.end(); i++) {
    const SCTDAL::SCT_Module *modObj = (*i)->get_module();
    if(modObj) {
      result.push_back(modObj->get_sn());
    }
  }

  return result;
}

std::list<std::string> ConfigurationDALImpl::listAllModules() {
  std::vector<const SCTDAL::SCT_Module *> modules;
  conf().get<SCTDAL::SCT_Module>(modules);

  std::list<std::string> result;
  for(SCTDAL::SCT_ModuleIterator i = modules.begin(); i!=modules.end(); i++) {
    result.push_back((*i)->get_sn());
  }

  return result;
}

std::list<std::string> ConfigurationDALImpl::listUnusedModules() {
  std::vector<const SCTDAL::SCT_Module *> modules;
#warning "Could make up a query?"
  conf().get<SCTDAL::SCT_Module>(modules);

  std::list<std::string> result;
  for(SCTDAL::SCT_ModuleIterator i = modules.begin(); i!=modules.end(); i++) {
    // Not attached to anything?
    const SCTDAL::SCT_Module *modObj = (*i);
    const SCTDAL::RodMUR *murObj = 0;
    bool foundMur = lookup->getSingleBackReference(*modObj, "module", murObj);
    if(!foundMur) {
      result.push_back((*i)->get_sn());
    }
  }

  return result;
}

ABCDModule ConfigurationDALImpl::getModuleConfig(const std::string module) {
  const SCTDAL::SCT_Module *moduleObj = lookup->getModule(module);

  if(!moduleObj) {
    throw SctConfiguration::ConfigurationException("No configuration for module " + module);
  }

  ABCDModule moduleConf;

  for(unsigned int i=0; i<sizeof(moduleConf); i++) {
    ((char*)&moduleConf)[i] = 0;
  }

  // Defaults
  moduleConf.mid=0xffffffff;
  strncpy(moduleConf.serialNumber, module.c_str(), 16);

  // What's stored in the configuration
  moduleConf.active = moduleObj->get_active();
  moduleConf.select = moduleObj->get_select();

  std::vector<const SCTDAL::SCT_Chip*> chips = moduleObj->get_Chips();

  for(SCTDAL::SCT_ChipIterator i = chips.begin(); i != chips.end(); i++) {
    const SCTDAL::SCT_Chip *chipOb = *i;

    int chipNum = chipOb->get_id();
    ABCDChip &currChip = moduleConf.chip[chipNum];

    currChip.address = chipOb->get_address();
    currChip.active = chipOb->get_active();

    currChip.basic.vthr = chipOb->get_vthr();
    currChip.basic.vcal = chipOb->get_vcal();
    currChip.basic.delay = chipOb->get_delay();
    currChip.basic.preamp = chipOb->get_preamp();
    currChip.basic.shaper = chipOb->get_shaper();

    currChip.caldata.rc_function = chipOb->get_rc_function_type();

    parseRcFunction(chipOb->get_rc_function(), currChip.caldata.rc_params);

    std::vector<short unsigned int> configs;
    configs = chipOb->get_config();
    currChip.basic.config.readoutMode = configs[0];
    currChip.basic.config.calibMode = configs[1];
    currChip.basic.config.trimRange = configs[2];
    currChip.basic.config.edgeDetect = configs[3];
    currChip.basic.config.mask = configs[4];
    currChip.basic.config.accumulate = configs[5];
    currChip.basic.config.inputBypass = configs[6];
    currChip.basic.config.outputBypass = configs[7];
    currChip.basic.config.master = configs[8];
    currChip.basic.config.end = configs[9];
    currChip.basic.config.feedThrough = configs[10];

    currChip.caldata.c_factor = chipOb->get_c_factor();
    currChip.target = (int)((chipOb->get_target() / 2.5)+0.5);

    std::vector<unsigned long> trims;
    trims = chipOb->get_trims();

    for(int cluster=0; cluster<16; cluster++) {
      for(int i=0; i<8; i++) {
	unsigned char val = (trims[cluster] >> ((7-i)*4)) & 0x0000000f;
	currChip.trim[cluster*8+i] = val;
      }
    }

    std::vector<unsigned long> masks = chipOb->get_mask();
    reverse(masks.begin(), masks.end());
    copy(masks.begin(), masks.end(), &currChip.basic.mask[0]);
  }

  return moduleConf;
}

short ConfigurationDALImpl::getModuleGroup(const std::string module) 
{
  const SCTDAL::RodModule *mapObj = 0;
#if 0
  const SCTDAL::SCT_Module *modObj = lookup->getModule(module);
  if(!modObj) {
    throw SctConfiguration::ConfigurationException("Can't find serial number to translate");
  }

  bool foundMap = lookup->getSingleBackReference(*modObj, "module", mapObj);
  if(!foundMap) {
    throw SctConfiguration::ConfigurationException("Can't reverse module link for translate");
  }
#else
  mapObj = lookup->queryViaRelationship<SCTDAL::RodModule>("module", "sn", "\"" + module + "\"");
  if(!mapObj) {
    throw SctConfiguration::ConfigurationException("Can't find module in ROD");
  }
#endif
  return mapObj->get_group();
}

unsigned int ConfigurationDALImpl::getModuleErrorMask(unsigned int partition, unsigned int crate, unsigned int rod,
						      unsigned int channel, unsigned int link) {
  unsigned int order = channel / 6;
  unsigned int id = (channel % 6) + 1;

  const SCTDAL::Rod *rodObj = lookup->getRod(partition, crate, rod);
  if(!rodObj) {
    throw SctConfiguration::ConfigurationException("No ROD for error mask");
  }
  const SCTDAL::RodMUR *murObj = lookup->getMurInRod(rodObj, order);
  if(!murObj) {
    throw SctConfiguration::ConfigurationException("No MUR for error mask");
  }

  // default
  unsigned int value = 0;

  const SCTDAL::Channel *channelObj = lookup->getChannelInRodMur(murObj, id);
  if(channelObj) {
    if(link == 0) {
      value = channelObj->get_stream0ErrMask();
    } else {
      value = channelObj->get_stream1ErrMask();
    }

  }

  return 0;
}

SctConfiguration::RodConfig ConfigurationDALImpl::getRodConfig(unsigned int partition, unsigned int crate, unsigned int rod) {
  const SCTDAL::Rod *rodObj = lookup->getRod(partition, crate, rod);
  if(!rodObj) {
    throw SctConfiguration::ConfigurationException("Crate not found in ROD lookup");
  }

  SctConfiguration::RodConfig rodConfig;
  rodConfig.baseAddress = rodObj->get_slot() << 24;
#warning "Put this in?"
  rodConfig.resetLevel = 0;

  std::vector<const SCTDAL::Slave *> slaves = rodObj->get_slaves();

  for(SCTDAL::SlaveIterator i = slaves.begin(); i!=slaves.end(); i++) {
    const SCTDAL::Slave * slave = *i;
    int id = slave->get_id();

    rodConfig.slaves[id].ipramFile = slave->get_ipramFile();
    rodConfig.slaves[id].extFile = slave->get_extFile();
    rodConfig.slaves[id].idramFile = slave->get_idramFile();
  }

  rodConfig.validMask = 0;

  return rodConfig;
}

SctConfiguration::SlaveConfig ConfigurationDALImpl::getDefaultSlaveConfig(char type) {
#warning "This should be in getRODConfig instead (though useful for editting)"
  std::cerr << "Undefined at " << __LINE__ << std::endl;

  SctConfiguration::SlaveConfig slaveConfig;
  return slaveConfig;
}

SctConfiguration::BOCChannelConfig ConfigurationDALImpl::getBOCConfig(unsigned int partition, unsigned int crate, unsigned int rod, 
								      unsigned int channel) {
  if(channel < 0 || channel > 47) {
    throw SctConfiguration::ConfigurationException("Channel not in range (0 to 47) on input to getBOCConfig");
  }

  unsigned int order = channel/6;
  unsigned int chan = (channel % 6) + 1;

  const SCTDAL::Rod *rodObj = lookup->getRod(partition, crate, rod);

  if(!rodObj) {
    throw SctConfiguration::ConfigurationException("No ROD for BOC config");
  }

  const SCTDAL::RodMUR *murObj = 0;
  std::vector<const SCTDAL::RodMUR*> murs = rodObj->get_rodMURs();
  for(SCTDAL::RodMURIterator i = murs.begin(); i!=murs.end(); i++) {
    if((*i)->get_order() == order) {
      murObj = *i;
    }
  }

  if(!murObj) {
    throw SctConfiguration::ConfigurationException("No MUR for BOC config");
  }

  const SCTDAL::Channel *channelObj = 0;
  std::vector<const SCTDAL::Channel*> channels = murObj->get_channels();
  for(SCTDAL::ChannelIterator i = channels.begin(); i!=channels.end(); i++) {
    if((*i)->get_id() == chan) {
      channelObj = *i;
    }
  }

  if(!channelObj) {
    throw SctConfiguration::ConfigurationException("No BOC config in MUR");
  }

  SctConfiguration::BOCChannelConfig chanConfig;
  chanConfig.current = channelObj->get_outputCurrent();
  chanConfig.delay = channelObj->get_outputDelay();
  chanConfig.markSpace = channelObj->get_outputMarkSpace();

  chanConfig.threshold0 = channelObj->get_stream0Threshold();
  chanConfig.delay0 = channelObj->get_stream0Delay();

  chanConfig.threshold1 = channelObj->get_stream1Threshold();
  chanConfig.delay1 = channelObj->get_stream1Delay();

  return chanConfig;
}


SctConfiguration::BOCGlobalConfig ConfigurationDALImpl::getBOCGlobalConfig(unsigned int partition, unsigned int crate, unsigned int rod) {
#warning "undefined"
  std::cerr << "Undefined getBOCGlobalConfig at " << __LINE__ << std::endl;

  SctConfiguration::BOCGlobalConfig config;
  config.validMask = 0;
  return config;
}

unsigned char *ConfigurationDALImpl::getFibreMappings(unsigned int partition, unsigned int crate, unsigned int rod) {
  const SCTDAL::Rod *rodObj = lookup->getRod(partition, crate, rod);

  if(!rodObj) {
    throw SctConfiguration::ConfigurationException("No ROD for fibre mappings");
  }

  unsigned char *result = new unsigned char[48*3];

  for(int mur = 0; mur < 8; mur ++) {
    const SCTDAL::RodMUR *murObj = 0;
    std::vector<const SCTDAL::RodMUR *> murs = rodObj->get_rodMURs();

    for(SCTDAL::RodMURIterator i = murs.begin(); i!=murs.end(); i++) {
      if((*i)->get_order() == mur) {
	murObj = *i;
      }
    }

    for(int chan = 0; chan < 6; chan++) {
      int channel = mur * 6 + chan;

      // Defaults
      int txFibre = channel;
      int rxFibre0 = channel * 2;
      int rxFibre1 = channel * 2 + 1;

      if(murObj) {
 	const SCTDAL::Channel *chanObj = 0;

	std::vector<const SCTDAL::Channel *> channels = murObj->get_channels();
	for(SCTDAL::ChannelIterator i = channels.begin(); i!=channels.end(); i++) {
	  if((*i)->get_id() == chan+1) {
	    chanObj = (*i);
	  }
	}
	if(chanObj) {
	  log << "Found non standard mappings at " << channel << "\n";
	  int tx, rx0, rx1;
	  tx = chanObj->get_outputFibre();
	  rx0 = chanObj->get_stream0Fibre();
	  rx1 = chanObj->get_stream1Fibre();
	  if(tx != 0x80) txFibre = tx;
	  if(rx0 != 0x80) rxFibre0 = rx0;
	  if(rx1 != 0x80) rxFibre1 = rx1;
	}
      }

      result[channel * 3 + 0] = txFibre;
      result[channel * 3 + 1] = rxFibre0;
      result[channel * 3 + 2] = rxFibre1;
    }
  }

  return result;
}

SctConfiguration::TimConfig ConfigurationDALImpl::getTimConfig(unsigned int partition, unsigned int crate) {
  const SCTDAL::Crate *crateObj = lookup->getCrate(partition, crate);
  if(!crateObj) {
    throw SctConfiguration::ConfigurationException("Crate not found in TIM lookup");
  }

  const SCTDAL::Tim *timObj = crateObj->get_tim();
  if(!timObj) {
    throw SctConfiguration::ConfigurationException("No TIM in crate");
  }

//   std::cerr << "Found TIM:\n";
//   timObj->print(2, true, std::cerr);
//   std::cerr << "In crate\n";
//   crateObj->print(1, true, std::cerr);

  SctConfiguration::TimConfig config;
  config.baseAddress = 0xd << 24;
  config.trigFrequency = timObj->get_triggerFrequency();
  config.resetFrequency = timObj->get_resetFrequency();

  config.validMask = 0;

  return config;
}

std::string ConfigurationDALImpl::printModuleConfig(const ABCDModule &conf) {
  std::ostringstream oss;
  
  SCTConfigCommonUtils::printModuleConfig(conf, oss);

  return oss.str();
}


SctConfiguration::MURType ConfigurationDALImpl::getMURType(unsigned int MUR) {
  const SCTDAL::GeographyMUR *murObj = lookup->getGeogMur(MUR);

  if(murObj == 0) {
    const SCTDAL::RodMUR *rodMur = lookup->getMur(MUR);
    if(rodMur == 0) {
      return SctConfiguration::UNKNOWN;
    } else {
      return SctConfiguration::UNMAPPED;
    }
  }

  const SCTDAL::Row *row = 0;

  bool success = lookup->getSingleBackReference(*murObj, "murs", row);

  if(success) {
    return SctConfiguration::BARREL;
  } else {
#warning "Need to check for endcap when schema supports it"
    // UNMAPPED should be for if its there as a ROD mur
    return SctConfiguration::UNMAPPED;
  }
}


unsigned int ConfigurationDALImpl::getFreeMurId() {
#warning "Very little used"
  std::cerr << "Undefined at " << __LINE__ << std::endl;

  return 0;
}

void ConfigurationDALImpl::translateToROD(unsigned int MUR, unsigned int module,
					  unsigned int &partition, unsigned int &crate, 
					  unsigned int &rod, unsigned int &channel) {
  const SCTDAL::RodMUR *murOb = lookup->getMur(MUR);
  if(!murOb) {
    throw SctConfiguration::ConfigurationException("No MUR in translateToROD");
  }
  channel = murOb->get_order() * 6 + module - 1;

  const SCTDAL::Rod *rodOb = 0;
  bool foundRod = lookup->getSingleBackReference(*murOb, "rodMURs", rodOb);
  if(!foundRod) {
    throw SctConfiguration::ConfigurationException("MUR not in ROD");
  }
  rod = rodOb->get_id();

  const SCTDAL::Crate *crateOb = 0;
  bool foundCrate = lookup->getSingleBackReference(*rodOb, "rods", crateOb);
  if(!foundCrate) {
    throw SctConfiguration::ConfigurationException("ROD not in crate");
  }
  crate = crateOb->get_id();

  const SCTDAL::Partition *partOb = 0;
  //  partOb = lookup->queryViaRelationship<SCTDAL::Partition>("crates", "id", boost::lexical_cast<std::string>(crate));
  bool foundPart = lookup->getSingleBackReference(*crateOb, "crates", partOb);
  if(!foundPart) {
    throw SctConfiguration::ConfigurationException("No partition link");
  }
  partition = partOb->get_id();
}

void ConfigurationDALImpl::translateToRROD(unsigned int MUR, unsigned int module,
					   unsigned int &partition, unsigned int &crate, 
					   unsigned int &rod, unsigned int &channel) {
  std::string sn;

  translateToSN(MUR, module, sn);

  const SCTDAL::RodRModule *rMapOb = lookup->queryViaRelationship<SCTDAL::RodRModule>("module", "sn", "\"" + sn + "\"");
  if(!rMapOb) {
    throw SctConfiguration::ConfigurationException("No link to RMUR");
  }

  const SCTDAL::RodMUR *murOb = 0;
  bool foundMUR = lookup->getSingleBackReference(*rMapOb, "rodRModules", murOb);
  if(!foundMUR) {
    throw SctConfiguration::ConfigurationException("No MUR in translateToRROD");
  }

  channel = murOb->get_order() * 6 + module - 1;

  const SCTDAL::Rod *rodOb = 0;
  bool foundROD = lookup->getSingleBackReference(*murOb, "rodMURs", rodOb);
  if(!foundROD) {
    throw SctConfiguration::ConfigurationException("No ROD link");
  }
  rod = rodOb->get_id();

  const SCTDAL::Crate *crateOb = 0;
  bool foundCrate = lookup->getSingleBackReference(*rodOb, "rods", crateOb);
  if(!foundCrate) {
    throw SctConfiguration::ConfigurationException("No crate link");
  }
  crate = crateOb->get_id();

  const SCTDAL::Partition *partOb = 0;
  bool foundPart = lookup->getSingleBackReference(*crateOb, "crates", partOb);
  if(!foundPart) {
    throw SctConfiguration::ConfigurationException("No partition link");
  }
  partition = partOb->get_id();
}

void ConfigurationDALImpl::translateToSN(unsigned int MUR, unsigned int module,
					 std::string &sn) {
  const SCTDAL::RodMUR *murOb = lookup->getMur(MUR);
  std::vector<const SCTDAL::RodModule *> modules = murOb->get_rodModules();

  bool found = false;

  for(SCTDAL::RodModuleIterator i = modules.begin(); i!=modules.end(); i++) {
    if((*i)->get_position() == module) {
      const SCTDAL::SCT_Module *modObj = (*i)->get_module();
      if(modObj) {
	sn = modObj->get_sn();
	found = true;
      }
#warning "Check duplicates"
    }
  }
  if(!found) {
    throw SctConfiguration::ConfigurationException("Module not found in MUR");
  }
}

void ConfigurationDALImpl::translateToRMUR(unsigned int MUR, unsigned int module,
					   unsigned int &RMUR, unsigned int &rmodule) {
  std::string sn;
  translateToSN(MUR, module, sn);

#if 1
#warning "But, this might pick up the RodModule link too?"
  const SCTDAL::RodRModule *mapObj = lookup->queryViaRelationship<SCTDAL::RodRModule>("module", "sn", "\"" + sn + "\"");
#else
  std::string query("(this (\"module\" some (\"sn\" \"" + sn + "\" =)))");

  log << "executing query: " << query << std::endl;

  std::vector<const SCTDAL::RodRModule *> maps;
  conf().get<SCTDAL::RodRModule>(maps, false, true, query);
#endif

  const SCTDAL::RodMUR *murObj = 0;
  bool foundMUR = lookup->getSingleBackReference(*mapObj, "rodRModules", murObj);
  if(!foundMUR) {
    throw SctConfiguration::ConfigurationException("Can't find reverse redundant MUR link for translate");
  }
  if(!murObj) {
    throw SctConfiguration::ConfigurationException("Oops in toRMUR!");
  }

  MUR = murObj->get_id();
  module = mapObj->get_position();
}

void ConfigurationDALImpl::translateToBarrel(unsigned int MUR, unsigned int module,
					     unsigned int &barrel, unsigned int &row, int &number) {
  const SCTDAL::GeographyMUR *murOb = lookup->getGeogMur(MUR);

  if(!murOb) {
    throw SctConfiguration::ConfigurationException("Requested MUR not mapped to barrel");
  }

  // Cast to make sure it doesn't comare pointers...
  if(murOb->get_position() == std::string("-1")) {
    number = -module;
  } else {
    number = module;
  }

  const SCTDAL::Row *rowOb = 0;
#if 0
  bool foundRow = lookup->getSingleBackReference(*murOb, "murs", rowOb);

  if(!foundRow) {
    throw SctConfiguration::ConfigurationException("ROD does't own up to MUR");
  }

  if(!rowOb) {
    throw SctConfiguration::ConfigurationException("getSingleBackReference returned null and true!");
  }
#else
  rowOb = lookup->queryViaRelationship<SCTDAL::Row>("murs", "id", "\"" + boost::lexical_cast<std::string>(murOb->get_id()) + "\"");
  if(!rowOb) {
    throw SctConfiguration::ConfigurationException("backReference failed");
  }
#endif

  row = rowOb->get_id();

#if 1
  const SCTDAL::Barrel *barrelOb = 0;
  bool success = lookup->getSingleBackReference(*rowOb, "rows", barrelOb);

  if(!success) {
    throw SctConfiguration::ConfigurationException("Row not in a barrel");
  }
  if(!barrelOb) {
    throw SctConfiguration::ConfigurationException("True and null");
  }
#else
  const SCTDAL::Barrel *barrelOb = lookup->queryViaRelationship<SCTDAL::Barrel>("rows", "id", "\"" + boost::lexical_cast<std::string>(row) + "\"");
  if(!barrelOb) {
    throw SctConfiguration::ConfigurationException("No Barrel for Row");
  }
#endif

  barrel = barrelOb->get_id();
}

void ConfigurationDALImpl::translateToEndcap(unsigned int MUR, unsigned int module,
					     int &disk, unsigned int &ring, unsigned int &number) {
#warning "undefined endcap"
  std::cerr << "Undefined at " << __LINE__ << std::endl;
}

void ConfigurationDALImpl::translateToPowerSupply(unsigned int MUR, unsigned int module,
                                                  unsigned int &partition, 
                                                  unsigned int &crate, unsigned int &channel) {
  const SCTDAL::PowerChannel *chanObj = lookup->getPowerChannelByMUR(MUR, module);

  if(!chanObj) {
    throw SctConfiguration::ConfigurationException("Module not mapped to power supply");
  }

  channel = chanObj->get_id();

  const SCTDAL::PowerCrate *crtOb = 0;
  bool foundCrate = lookup->getSingleBackReference(*chanObj, "channels", crtOb);
  if(!foundCrate) {
    throw SctConfiguration::ConfigurationException("Channel not in crate");
  }

  crate = crtOb->get_id();

  // Fixed
  partition = 0;
}

void ConfigurationDALImpl::translateFromROD(unsigned int partition, unsigned int crate, 
					    unsigned int rod, unsigned int channel,
					    unsigned int &MUR, unsigned int &module) {
  unsigned int order = channel / 6;
  module = (channel % 6) + 1;

  const SCTDAL::Rod *rodObj = lookup->getRod(partition, crate, rod);

  std::vector<const SCTDAL::RodMUR*> murs = rodObj->get_rodMURs();

  for(SCTDAL::RodMURIterator i = murs.begin(); i!= murs.end(); i++) {
    if((*i)->get_order() == order) {
      MUR = (*i)->get_id();
#warning "Check duplicates"
    }
  }
}

void ConfigurationDALImpl::translateFromRROD(unsigned int partition, unsigned int crate,
					     unsigned int rod, unsigned int channel,
					     unsigned int &MUR, unsigned int &module) {
  unsigned int order = channel / 6;
  unsigned int id = (channel % 6) + 1;

  const SCTDAL::Rod *rodObj = lookup->getRod(partition, crate, rod);
  if(!rodObj) {
    throw SctConfiguration::ConfigurationException("No Rod for from RROD");
  }
  const SCTDAL::RodMUR *murObj = lookup->getMurInRod(rodObj, order);
  if(!murObj) {
    throw SctConfiguration::ConfigurationException("No MUR for from RROD");
  }
  const SCTDAL::RodRModule *mapObj = lookup->getRModuleInRodMur(murObj, id);
  if(!mapObj) {
    throw SctConfiguration::ConfigurationException("No mapping for from RROD");
  }
  const SCTDAL::SCT_Module *modObj = mapObj->get_module();
  if(!modObj) {
    throw SctConfiguration::ConfigurationException("No module mapped");
  }
  std::string sn = modObj->get_sn();
  translateFromSN(sn, MUR, module);
}

void ConfigurationDALImpl::translateFromSN(const std::string sn,
					   unsigned int &MUR, unsigned int &module) {
  const SCTDAL::SCT_Module *modObj = lookup->getModule(sn);

  if(!modObj) {
    throw SctConfiguration::ConfigurationException("Can't find serial number to translate");
  }

  const SCTDAL::RodModule *mapObj = 0;
#if 0
  bool foundModuleMap = lookup->getSingleBackReference(*modObj, "module", mapObj);
  if(!foundModuleMap) {
    throw SctConfiguration::ConfigurationException("Can't find reverse reference for translate");
  }
  if(!mapObj) {
    throw SctConfiguration::ConfigurationException("getSingleBackReference returned true and NULL (map)");
  }
#else 
  mapObj = lookup->queryViaRelationship<SCTDAL::RodModule>("module", "sn", "\"" + sn + "\"");
  if(!mapObj) {
    throw SctConfiguration::ConfigurationException("Can't get from sn to MUR");
  }
#endif

  const SCTDAL::RodMUR *murObj = 0;
  bool foundMUR = lookup->getSingleBackReference(*mapObj, "rodModules", murObj);
  if(!foundMUR) {
    throw SctConfiguration::ConfigurationException("Can't find reverse MUR link for translate");
  }
  if(!murObj) {
    throw SctConfiguration::ConfigurationException("Oops in from SN!");
  }

  MUR = murObj->get_id();
  module = mapObj->get_position();
}

void ConfigurationDALImpl::translateFromRMUR(unsigned int RMUR, unsigned int rmodule,
					     unsigned int &MUR, unsigned int &module) {
  const SCTDAL::RodMUR *murObj = lookup->getMur(RMUR);

  const SCTDAL::RodRModule *mapObj = lookup->getRModuleInRodMur(murObj, rmodule);

  std::string sn = mapObj->get_module()->get_sn();

  translateFromSN(sn, MUR, module);
}

void ConfigurationDALImpl::translateFromBarrel(unsigned int barrel, unsigned int row, int number,
					       unsigned int &MUR, unsigned int &module) {
  const SCTDAL::Row *rowObj = lookup->getBarrelRow(barrel, row);

  std::string position;
  if(number < 0) {
    position = "-1";
    module = -number;
  } else {
    position = "1";
    module = number;
  }

  std::vector<const SCTDAL::GeographyMUR *> murs = rowObj->get_murs();
  const SCTDAL::GeographyMUR *murObj = 0;

  for(SCTDAL::GeographyMURIterator i = murs.begin(); i != murs.end(); i++) {
    if((*i)->get_position() == position) {
      murObj = *i;
    }
  }

  MUR = murObj->get_id();
}

void ConfigurationDALImpl::translateFromEndcap(int disk, unsigned int ring, unsigned int number,
					       unsigned int &MUR, unsigned int &module) {

#warning "undefined endcap"
  std::cerr << "Undefined at " << __LINE__ << std::endl;
}

void ConfigurationDALImpl::translateFromPowerSupply(unsigned int partition, 
                                                    unsigned int crate, unsigned int channel, 
                                                    unsigned int &MUR, unsigned int &module) {
  const SCTDAL::PowerChannel *chanObj = lookup->getPowerChannel(crate, channel);

  MUR = chanObj->get_MUR();
  module = chanObj->get_module();
}

void ConfigurationDALImpl::getMapMURROD(unsigned int MUR, unsigned int &partition, unsigned int &crate, unsigned int &rod, unsigned int &order) {
#warning "Editing only"
  std::cerr << "Undefined at " << __LINE__ << std::endl;
}

void ConfigurationDALImpl::getMapRODMUR(unsigned int partition, unsigned int crate, unsigned int rod, unsigned int order, unsigned int &MUR) {
#warning "Editing only"
  std::cerr << "Undefined at " << __LINE__ << std::endl;
}

void ConfigurationDALImpl::getMapMURPhysical(unsigned int MUR, SctConfiguration::MURType &type, int &plane, int &section, int &position) {
#warning "Editing only"
  std::cerr << "Undefined at " << __LINE__ << std::endl;
}

float ConfigurationDALImpl::getDefaultPowerParam(std::string state, std::string name, std::string type) {
#warning "Editing only"
  std::cerr << "Undefined at " << __LINE__ << std::endl;

  return 0.0;
}

float ConfigurationDALImpl::getHVCardParam(unsigned int crate, unsigned int card, std::string name, std::string type) {
  const SCTDAL::PowerCrate *crateObj = lookup->getPowerCrate(crate);
  if(!crateObj) {
    throw SctConfiguration::ConfigurationException("Crate does not exist");
  }

  const SCTDAL::PowerCard *cardObj = lookup->getPowerCardInCrate(crateObj, card, "HV");
  if(!cardObj) {
    throw SctConfiguration::ConfigurationException("Card does not exist");
  }

#warning "Fininsh this off!!!"

  return 0.0;
}

float ConfigurationDALImpl::getLVCardParam(unsigned int crate, unsigned int card, std::string name, std::string type) {
#warning "undefined"
  std::cerr << "Undefined at " << __LINE__ << std::endl;

  return 0.0;
}

float ConfigurationDALImpl::getPowerParam(unsigned int MUR, unsigned int number, std::string state, std::string name, std::string type) {
  const SCTDAL::PowerChannel *chanObj = lookup->getPowerChannelByMUR(MUR, number);
  if(!chanObj) {
    throw SctConfiguration::ConfigurationException("MUR/module pair not mapped to power");
  }

  const SCTDAL::PowerParam *paramObj = lookup->getPowerParamInChannel(chanObj, state, name);

  float value = 0.0;

  if(!paramObj) {
    // Lookup in defaults
    std::vector<const SCTDAL::Power *> powerList;
    conf().get<SCTDAL::Power>(powerList);

    std::vector<const SCTDAL::PowerParam *> defaults;
    if(powerList.size() == 1) {
      defaults = powerList.front()->get_defaults();
    }

    for(SCTDAL::PowerParamIterator i = defaults.begin(); i!=defaults.end(); i++) {
      if((*i)->get_state() == state && (*i)->get_name() == name) {
	paramObj = *i;
      }
    }
  }

  if(!paramObj) {
    throw SctConfiguration::ConfigurationException("Can't find appropriate power parameter");
  }

  if(type == "loAlarm") {
    value = paramObj->get_loAlarm();
  } else if(type == "loWarn") {
    value = paramObj->get_loWarn();
  } else if(type == "value") {
    value = paramObj->get_value();
  } else if(type == "hiWarn") {
    value = paramObj->get_hiWarn();
  } else if(type == "hiAlarm") {
    value = paramObj->get_hiAlarm();
  } else {
    throw SctConfiguration::ConfigurationException("Unknown parameter type");
  }

  return value;
}

float ConfigurationDALImpl::getPowerParamCC(unsigned int crate, unsigned int channel, std::string state, std::string name, std::string type) {
#warning "undefined"
  std::cerr << "Undefined at " << __LINE__ << std::endl;

  return 0.0;
}

std::list<std::string> ConfigurationDALImpl::listDefaultPowerParams() {
#warning "Little used, probably before listDCSChannelParams"
  std::cerr << "Undefined at " << __LINE__ << std::endl;

  std::list<std::string> result;
  return result;
}

float ConfigurationDALImpl::getCratePowerParam(unsigned int crate, std::string name) {
#warning "Obsolete?"
  std::cerr << "Undefined at " << __LINE__ << std::endl;

  return 0.0;
}

std::list<std::string> ConfigurationDALImpl::listDCSChannelParams() {
//   log << "listDCSChannelParams\n";

  return SCTConfigCommonUtils::listDCSChannelParams();
}

std::list<std::string> ConfigurationDALImpl::listDCSCardParams() {
//   log << "listDCSCardParams\n";

  return SCTConfigCommonUtils::listDCSCardParams();
}

std::list<std::string> ConfigurationDALImpl::listDCSCrateParams() {
//   log << "listDCSCrateParams\n";

  return SCTConfigCommonUtils::listDCSCrateParams();
}

void ConfigurationDALImpl::useISForChanges(bool valid) {
#warning "undefined"
  std::cerr << "Undefined useISForChanges at " << __LINE__ << std::endl;
}

void ConfigurationDALImpl::useMRSForMessages(bool valid) {
#warning "undefined"
  std::cerr << "Undefined at " << __LINE__ << std::endl;
}


void ConfigurationDALImpl::loadConfiguration(const std::string &filename) {
#warning "undefined"
  std::cerr << "Undefined at " << __LINE__ << std::endl;
}

std::string ConfigurationDALImpl::getConfigurationFileName() {
  return lastFilename;
}

void ConfigurationDALImpl::saveConfiguration(const std::string filename) {
  std::string command = "cp " + lastFilename + " " + filename;

  std::cerr << "Executing " << command << std::endl;
  system(command.c_str());
#warning "Is this the right thing to do here???"
}

void ConfigurationDALImpl::saveModuleConfiguration(const std::string modsn, const std::string filename) {
#warning "Where is this used???"
  std::cerr << "Undefined at " << __LINE__ << std::endl;
}

std::string ConfigurationDALImpl::getModuleConfigurationString(const std::string modsn, const ABCDModule config) {
#warning "Where is this used?"
  return "<Dummy>xml file while I think what to do...<oops/> What is this used for by the way?</Dummy>";
}

void ConfigurationDALImpl::writePowerSupplyConfiguration(std::string filename) {
#warning "Obsolete"
  std::cerr << "Undefined at " << __LINE__ << std::endl;
}
